from .. import Interface

central = Interface('https://central.xnat.org', 'nosetests', 'nosetests')

def test_users():
    assert isinstance(central.manage.users(), list)

def test_user_firstname():
    assert central.manage.users.firstname('nosetests') == 'Yannick'

def test_user_lastname():
    assert central.manage.users.lastname('nosetests') == 'Schwartz'

def test_user_email():
    assert central.manage.users.email('nosetests') == \
        'yannick.schwartz@gmail.com'

def test_user_id():
    assert central.manage.users.id('nosetests') == '204'

def test_project_users():
    assert isinstance(central.select.project('nosetests').users(), list)

def test_project_owners():
    assert isinstance(central.select.project('nosetests').owners(), list)

def test_project_members():
    assert isinstance(central.select.project('nosetests').members(), list)

def test_project_collaborators():
    assert isinstance(central.select.project('nosetests').collaborators(), 
                      list
                      )

def test_project_user_role():
    assert central.select.project('nosetests'
                                  ).user_role('nosetests') == 'owner'

def test_add_remove_user():
    central.select.project('nosetests').add_user('schwarty', 'collaborator')
    assert 'schwarty' in central.select.project('nosetests').collaborators()
    central.select.project('nosetests').remove_user('schwarty')
    assert 'schwarty' not in central.select.project('nosetests'
                                                    ).collaborators()

def test_project_accessibility():
    assert central.select.project('nosetests').accessibility() in \
                                        ['public', 'protected', 'private']
    central.select.project('nosetests').set_accessibility('private')
    assert central.select.project('nosetests').accessibility() == 'private'
    central.select.project('nosetests').set_accessibility('protected')
    assert central.select.project('nosetests').accessibility() == 'protected'

def test_project_prearchive_code():
    pass

def test_project_quarantine_code():
    pass

def test_current_arc():
    pass




_boundary = '----------ThIs_Is_tHe_bouNdaRY_$'
_crlf = '\r\n'


def file_message(content, content_type, path, name):
    body = []
    body.append('--' + _boundary)
    body.append('Content-Disposition: form-data; '
                'name="%s"; filename="%s"' % (path, name)
                )
    body.append('Content-Type: %s' % content_type)
    body.append('')

    body.append(content)
    body.append('--' + _boundary + '--')
    body.append('')
    body = _crlf.join(body)
    content_type = 'multipart/form-data; boundary=%s' % _boundary

    return body, content_type




# REST collection resources tree
resources_tree = {
    'projects'        :['subjects', 'resources'],
    'subjects'        :['experiments', 'resources'],
    'experiments'     :['assessors', 'reconstructions', 'scans', 'resources'],
    'assessors'       :['resources', 'in_resources','out_resources'],
    'reconstructions' :['in_resources','out_resources'],
    'scans'           :['resources'],
    'resources'       :['files'],
    'files'           :[],
    'in_resources'    :['files'],
    'in_files'        :[],
    'out_resources'   :['files'],
    'out_files'       :[],
    }

prearc_tree = {'projects'   :['scans'],
               'scans'      :['resources'],
               'resources'  :['files']
               }
# REST resources that are not natively supported
extra_resources_tree = {'projects':['assessors', 'scans', 'reconstructions'],
                        'subjects':['assessors', 'scans', 'reconstructions'],
                        }

# REST <Python to URI> translation table 
rest_translation = {'in_resources':'in/resources',
                    'in_files':'in/files',
                    'out_resources':'out/resources',
                    'out_files':'out/files',
                    'in_resource':'in/resource',
                    'in_file':'in/file',
                    'out_resource':'out/resource',
                    'out_file':'out/file',
                    }



# REST json format <id_header, label_header>
json = {'projects':['ID', 'ID'],
        'subjects':['ID', 'label'],
        'experiments':['ID', 'label'],
        'assessors':['ID', 'label'],
        'reconstructions':['ID', 'label'],
        'scans':['ID', 'ID'],
        'resources':['xnat_abstractresource_id', 'label'],
        'out_resources':['xnat_abstractresource_id', 'label'],
        'in_resources':['xnat_abstractresource_id', 'label'],
        'files':['path', 'path'],
        }

resources_singular = [key.rsplit('s', 1)[0] for key in resources_tree.keys()]
resources_plural   = resources_tree.keys()
resources_types    = resources_singular + list(resources_plural)

default_datatypes = {'projects':'xnat:projectData',
                     'subjects':'xnat:subjectData',
                     'experiments':'xnat:mrSessionData',
                     'assessors':'xnat:mrAssessorData',
                     'reconstructions':'xnat:reconstructedImageData',
                     'scans':'xnat:mrScanData',
                     'resources':None,
                     'in_resources':None,
                     'out_resources':None,
                     }

def datatype_attributes(root, datatype):
    def _iterchildren(node, pathsofar):
        elements = []

        for child in node.iterchildren():
            if isinstance(child.tag, basestring) \
                and child.tag.split('}')[1] == 'element':
                    elements.append('%s/%s'%(pathsofar, child.get('name')))
                    elements.extend(_iterchildren(child, '%s/%s'%
                                        (pathsofar, child.get('name')))
                                    )

            elif isinstance(child.tag, basestring) \
                and child.tag.split('}')[1] == 'attribute':
                    elements.append('%s/%s'%(pathsofar, child.get('name')))

            elif isinstance(child.tag, basestring) \
                    and child.tag.split('}')[1] == 'extension':
                
                ct_xpath = "/xs:schema/xs:complexType[@name='%s']"% \
                    child.get('base').split(':')[1]

                for complex_type in node.getroottree(
                    ).xpath(ct_xpath, namespaces=child.nsmap):

                    same = False

                    for ancestor in child.iterancestors():
                        if ancestor.get('name') == \
                                child.get('base').split(':')[1]:
                            same = True
                            break

                    if not same:
                        elements.extend(_iterchildren(complex_type, 
                                                      pathsofar)
                                        )

                    elements.extend(_iterchildren(child, pathsofar))

            else:
                elements.extend(_iterchildren(child, pathsofar))

        return elements

    ct_xpath = "/xs:schema/xs:complexType[@name='%s']" % \
        datatype.split(':')[1]

    attributes = []

    for complex_type in root.xpath(ct_xpath, namespaces=root.nsmap):
        for child in complex_type.iterchildren():                
            attributes.extend(_iterchildren(child, datatype))

    return attributes

def datatypes(root):
    nsmap = get_nsmap(root)

    return [element.get('type')
            for element in root.xpath('/xs:schema/xs:element', 
                                      namespaces=nsmap)
            ]

def get_nsmap(node):
    nsmap = node.nsmap
    none_ns = node.nsmap.get(None)

    if none_ns != None:
        nsmap[none_ns.rsplit('/', 1)[1]] = none_ns
        del nsmap[None]

    return nsmap

def class_name(self):
    """
    Return the name of this class without qualification.
    eg. If the class name is "x.y.class" return only "class"
    """
    return self.__class__.__name__.split('.')[-1]



import re

import httplib2
from lxml import etree

# parsing functions

def is_xnat_error(message):
    
    return message.startswith('<!DOCTYPE') or message.startswith('<html>')

def parse_error_message(message):
    try:
        if message.startswith('<html>'):
            message_tree = etree.XML(message)
            error_tag = message_tree.find('.//h3')
            if error_tag:
                error = error_tag.xpath("string()")
        elif message.startswith('<!DOCTYPE'):
            message_tree = etree.XML(message)
            error_tag = message_tree.find('.//title')
            if error_tag and 'Not Found' in error_tag.xpath("string()"):
                error = error_tag.xpath("string()")
            else:
                error_tag = message_tree.find('.//h1')
                if error_tag :
                    error = error_tag.xpath("string()")
        else:
            error = message

    except Exception, e:
        error = message
    finally:
        return error

def parse_put_error_message(message):
    error = parse_error_message(message)

    required_fields = []

    if error:
        for line in error.split('\n'):

            try:
                datatype_name = re.findall("\'.*?\'",line)[0].strip('\'')
                element_name = re.findall("\'.*?\'",line
                                          )[1].rsplit(':', 1)[1].strip('}\'')

                required_fields.append((datatype_name, element_name))
            except:
                continue

    return required_fields

def catch_error(msg_or_exception):

    # handle errors returned by the xnat server
    if isinstance(msg_or_exception, (str, unicode)):
        # parse the message
        msg = msg_or_exception
        error = parse_error_message(msg)
            
        # choose the exception
        if error == 'The request requires user authentication':
            raise OperationalError('Authentication failed')
        elif 'Not Found' in error:
            raise OperationalError('Connection failed')
        else:
            raise DatabaseError(error)

    # handle other errors, raised for instance by the http layer
    else:
        if isinstance(msg_or_exception, httplib2.ServerNotFoundError):
            raise OperationalError('Connection failed')
        else:
            raise DatabaseError(str(msg_or_exception))


# Exceptions as defined in PEP-249, the module treats errors using thoses
# classes following as closely as possible the original definitions.


# http://python3porting.com/differences.html#standarderror
try:
    class Warning(StandardError):
        pass

    class Error(StandardError):
        pass
except NameError:
    class Warning(Exception):
        pass

    class Error(Exception):
        pass


class InterfaceError(Error):
    pass

class DatabaseError(Error):
    pass

class DataError(DatabaseError):
    pass

class OperationalError(DatabaseError):
    pass

class IntegrityError(DatabaseError):
    pass

class InternalError(DatabaseError):
    pass

class ProgrammingError(DatabaseError):
    pass

class NotSupportedError(DatabaseError):
    pass



import csv
import copy
from fnmatch import fnmatch
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import json

# jdata is a list of dicts

def join_tables(join_column, jdata, *jtables):
    indexes = []

    for jtable in [jdata]+list(jtables):
        if isinstance(jtable, dict):
            jtable = [jtable]
        index = {}
        [index.setdefault(entry[join_column], entry) for entry in jtable]
        indexes.append(index)

    merged_jdata = []
    for join_id in indexes[0].keys():
        for index in indexes[1:]:
            indexes[0][join_id].update(index[join_id])
            merged_jdata.append(indexes[0][join_id])

    return merged_jdata

def get_column(jdata, col, val_pattern='*'):
    if isinstance(jdata, dict):
        jdata = [jdata]

    if val_pattern == '*':
        return [entry[col] for entry in jdata if entry.has_key(col)]
    else:
        return [entry[col] for entry in jdata 
                if fnmatch(entry.get(col), val_pattern)
                ]

def get_where(jdata, *args, **kwargs):
    if isinstance(jdata, dict):
        jdata = [jdata]

    match = []

    for entry in jdata:
        match_args = all([arg in entry.keys() or arg in entry.values() 
                          for arg in args
                          ])

        match_kwargs = all([entry[key] == kwargs[key] 
                            for key in kwargs.keys()
                            ])

        if match_args and match_kwargs:
            match.append(entry)

    return match

def get_where_not(jdata, *args, **kwargs):
    if isinstance(jdata, dict):
        jdata = [jdata]

    match = []

    for entry in jdata:
        match_args = all([arg in entry.keys() or arg in entry.values() 
                          for arg in args
                          ])

        match_kwargs = all([entry[key] == kwargs[key] 
                            for key in kwargs.keys()
                            ])

        if not match_args and not match_kwargs:
            match.append(entry)

    return match

def get_headers(jdata):
    if isinstance(jdata, dict):
        jdata = [jdata]
    return [] if jdata == [] else jdata[0].keys()

def get_selection(jdata, columns):
    if isinstance(jdata, dict):
        jdata = [jdata]

    sub_table = copy.deepcopy(jdata)

    rmcols = set(get_headers(jdata)).difference(columns)

    for entry in sub_table:
        for col in rmcols:
            if entry.has_key(col):
                del entry[col]

    return sub_table

def csv_to_json(csv_str):
    csv_reader = csv.reader(StringIO(csv_str), delimiter=',', quotechar='"')
    headers = csv_reader.next()

    return [dict(zip(headers, entry)) for entry in csv_reader]


class JsonTable(object):
    """ Wrapper around a list of dictionnaries to provide utility functions.
    """
    def __init__(self, jdata, order_by=[]):
        self.data = jdata
        self.order_by = order_by

    def __repr__(self):
        # if len(self.data) == 0:
        #     return '[]'
        # elif len(self.data) == 1:
        #     return str(self.data[0])
        
        if len(self.headers()) <= 5:
            _headers = ','.join(self.headers())
        else:
            _headers = '%s ... %s' % (','.join(self.headers()[:2]), 
                                      ','.join(self.headers()[-2:])
                                      )

        return '<JsonTable %s:%s> %s' % (
            len(self), len(self.headers()), _headers
            )
    
        # return ('[%s\n .\n .\n . \n%s]\n\n'
        #         '------------\n'
        #         '%s rows\n'
        #         '%s columns\n'
        #         '%s characters') % (str(self.data[0]), 
        #                             str(self.data[-1]),
        #                             len(self), 
        #                             len(self.headers()),
        #                             len(self.dumps_csv()) 
        #                             )

    def __str__(self):
        return self.dumps_csv()

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, name):
        if isinstance(name, (str, unicode)):
            return self.get(name)
        elif isinstance(name, int):
            return self.__class__([self.data[name]], self.order_by)
        elif isinstance(name, list):
            return self.select(name)

    def __getslice__(self, i, j):
        return self.__class__(self.data[i:j], self.order_by)

    def join(self, join_column, *jtables):
        """ Join jsontables with a common column.

            Parameters
            ----------
            join_column: string
                The name or header of the join column.
            jtables: *args
                Other  jtables.
        """
        return self.__class__(
            join_tables(join_column, self.data, 
                        *[jtable.data for jtable in jtables]),
            self.order_by
            )

    def has_header(self, name):
        return name in self.headers()

    def headers(self):
        """ Returns the headers of the object.
        """
        return get_headers(self.data)

    def get(self, col, val_pattern='*', always_list=False):
        """ Gets a single column value.

            Parameters
            ----------
            col: string
                The column name
            val_pattern: string
                Enable a filter on the values to be returned.
            always_list: boolean
                If only a single value is to be returned - i.e. there
                is only on element in the list of dicts or there is only
                one match against the value filter - is can be returned
                within a list (with True) or not (default).
        """
        res = get_column(self.data, col, val_pattern)
        if always_list:
            return res
        if len(self.data) == 1:
            return res[0]
        return res

    def where(self, *args, **kwargs):
        """ Filters the object.
        
            Paramaters
            ----------
            args:
                Value must be matched in the key or the value of an entry.
            kwargs:
                Value for a specific key must be matched in an entry.

            Returns
            -------
            A :class:`JsonTable` containing the matches.
        """
        return self.__class__(get_where(self.data, *args, **kwargs), 
                              self.order_by
                              )

    def where_not(self, *args, **kwargs):
        """ Filters the object. Conditions must not be matched.
        
            Paramaters
            ----------
            args:
                Value must not be matched in the key or the value of an 
                entry.
            kwargs:
                Value for a specific key must not be matched in an entry.

            Returns
            -------
            A :class:`JsonTable` containing the not matches.
        """
        return self.__class__(get_where_not(self.data, *args, **kwargs), 
                              self.order_by
                              )

    def select(self, columns):
        """ Select only some columns of interest.

            Returns
            -------
            A :class:`JsonTable` with the selected columns.
        """
        return self.__class__(get_selection(self.data, columns), 
                              self.order_by
                              )

    def dump_csv(self, dest, delimiter=','):
        """ Dumps the object content in a csv file format.

            Parameters
            ----------
            dest: string
                Destination file path.
            delimiter: char
                Character to separate values in the csv file.
        """
        fd = open(dest, 'w')
        fd.write(self.dumps_csv(delimiter))
        fd.close()

    def dumps_csv(self, delimiter=','):
        str_buffer = StringIO()
        csv_writer = csv.writer(str_buffer, delimiter=delimiter, 
                                quotechar='"', quoting=csv.QUOTE_MINIMAL)

        for entry in self.as_list():
            csv_writer.writerow(entry)

        return str_buffer.getvalue()

    def dump_json(self, dest):
        fd = open(dest, 'w')
        fd.write(self.dumps_json())
        fd.close()

    def dumps_json(self):
        return json.dumps(self.data)

    def as_list(self):
        table = [[]]

        for header in self.order_by:
            if header in self.headers():
                table[0].append(header)
        for header in self.headers():
            if header not in self.order_by:
                table[0].append(header)

        for entry in self.data:
            row = []
            for header in self.order_by:
                if entry.has_key(header):
                    row.append(entry.get(header))
            for header in self.headers():
                if header not in self.order_by:
                    row.append(entry.get(header))
            table.append(row)
        
        return table

    def items(self):
        table = []
    
        for entry in self.data:
            row = ()
            for header in self.order_by:
                if entry.has_key(header):
                    row += (entry.get(header), )
            for header in self.headers():
                if header not in self.order_by:
                    row += (entry.get(header), )
            table.append(row)
        
        return table




from __future__ import with_statement

import os
import platform
import ctypes
import glob
import hashlib
import re
import time
import shutil
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


_platform = platform.system()

DEBUG = False


def md5name(key):
    """ Generates a unique path to store server responses.
    """
    return hashlib.md5(key).hexdigest()

def bytes_to_human(size, unit):
    """ Returns a more human readable version of a size in bytes.
    """

    if unit == 'mega':
        return float(size) / 1024 ** 2
    elif unit == 'giga':
        return float(size) / 1024 ** 3
    else:
        return size

def memstr_to_bytes(text):
    """ Convert a memory text to its value in kilobytes.
    """
    kilo = 1024**2
    units = dict(K=1, M=kilo, G=kilo**2)
    try:
        size = int(units[text[-1]]*float(text[:-1]))
    except (KeyError, ValueError):
        raise ValueError(
                "Invalid literal for size give: %s (type %s) should be "
                "alike '10G', '500M', '50K'." % (text, type(text))
                )
    return size


class HTCache(object):
    def __init__(self, cachedir, interface, safe=md5name):
        """
            Parameters
            ----------
            cachedir: string
                The cache path.
            interface:
                :class:`Interface` Object
            safe: callable
                The function used to generate the responses cache paths.
        """
        self._intf = interface

        self.cache = cachedir
        self.safe = safe

        self._cachepath = None

        if not os.path.exists(cachedir):
            os.makedirs(cachedir)

    def get(self, key):
        retval = None

        _cachepath = os.path.join(self.cache, self.safe(key))

        if DEBUG:
            print('cache get:', key,)
            print('\n\t', _cachepath)

        try:
            f = file('%s.headers' % _cachepath, "rb")
            retval = f.read()
            f.close()

            f = file(_cachepath, "rb")
            retval += f.read()
            f.close()
        except IOError:
            try:
                f = file('%s.alt' % _cachepath, "rb")
                _altpath = f.read()
                f.close()

                f = file(_altpath, "rb")
                retval += f.read()
                f.close()
            except:
                return

        return retval

    def set(self, key, value):
        """ Sets cache entry.
        """

        _fakepath = '%s.alt' % os.path.join(self.cache, self.safe(key))
        _headerpath = '%s.headers' % os.path.join(self.cache, self.safe(key))
        _cachepath = os.path.join(self.cache, self.safe(key))

        if self._cachepath is not None: # when using custom path
            if _cachepath != self._cachepath and os.path.exists(_cachepath):
                os.remove(_cachepath) # remove default file if exists

            if os.path.exists(_fakepath):
                f = file(_fakepath, "rb")
                _altpath = f.read()
                f.close()

                if _altpath != self._cachepath and os.path.exists(_altpath):
                    # remove old custom file if different location
                    os.remove(_altpath)

            _cachepath = self._cachepath

            if DEBUG:
                print('cache set custom:', key)
                print('\n\t', _cachepath)

            f = open(_fakepath, 'w')
            f.write(_cachepath)
            f.close()

            self._cachepath = None

        else:                   # when using default cache path
            if os.path.exists(_fakepath): # if alternate location exists
                f = file(_fakepath, "rb")
                _altpath = f.read()
                f.close()

                if _altpath != self._cachepath and os.path.exists(_altpath):
                    os.remove(_altpath) # remove actual file

                os.remove(_fakepath) # remove pointer file

            if DEBUG:
                print('cache set default:', key)
                print('\n\t', _cachepath)

        header = ''
        value = StringIO(value)

        # avoid checking disk status each time
        if time.gmtime(time.time())[5] % 10 == 0:
            disk_status = self._intf.cache.disk_ready(_cachepath)

            if not disk_status[0] and self._intf.cache._warn:
                print('Warning: %s is %.2f%% full') % \
                    (os.path.dirname(_cachepath), disk_status[1])

        while len(header) < 4 or header[-4:] != '\r\n\r\n':
            header += value.read(1)

        content = value.read()

        f = file(_headerpath, "wb")
        f.write(header)
        f.close()

        f = file(_cachepath, "wb")
        f.write(content)
        f.close()

    def preset(self, path):
        """ Sets and forces a path for the next entry to be set.

            .. note::
                 Basically it is a way to trick the cache mechanism
                 into using something else than the default path to
                 store entries.
        """
        if self._intf._mode != 'offline':
            self._cachepath = path

    def delete(self, key):
        """ Deletes the entry.
        """

        if DEBUG:
            print('cache del:', key)

        _cachepath = os.path.join(self.cache, self.safe(key))
        _fakepath = '%s.alt' % _cachepath
        _headerpath = '%s.headers' % _cachepath

        if os.path.exists(_fakepath):
            f = file(_fakepath, "rb")
            _altpath = f.read()
            f.close()

            os.remove(_fakepath)
            if os.path.exists(_altpath):
                os.remove(_altpath)

        if os.path.exists(_cachepath):
            os.remove(_cachepath)

        if os.path.exists(_headerpath):
            os.remove(_headerpath)

    def get_diskpath(self, key, force_default=False):
        """ Gets the disk path where the entry is stored (default path
            or not)
        """
        _cachepath = os.path.join(self.cache, self.safe(key))
        _fakepath = '%s.alt' % _cachepath

        if os.path.exists(_fakepath) and not force_default:
            f = file(_fakepath, "rb")
            _altpath = f.read()
            f.close()
            return _altpath
        else:
            return _cachepath


class CacheManager(object):
    """ Management interface for the cache.

        It provides a few methods to::
            - evaluate the size a the cache
            - check if there is space left on the disk
            - clear the cache
            - define cache usage parameters
    """
    def __init__(self, interface):
        """
            Parameters
            ----------
            interface:
                :class:`Interface` Object

        """
        self._intf = interface
        self._cache = interface._http.cache
        self._warn = True

    def enable_warnings(self, toggle=True):
        self._warn = toggle

    def clear(self):
        """ Clears all files tracked by pyxnat.
        """
        for _fakepath in glob.iglob('%s/*.alt' % self._cache.cache):
            f = file(_fakepath, "rb")
            _altpath = f.read()
            f.close()

            if os.path.exists(_altpath):
                os.remove(_altpath)

        shutil.rmtree(self._cache.cache)
        os.makedirs(self._cache.cache)

    def size(self, unit='bytes'):
        """ Returns the amount of space taken by the cache.

            Parameters
            ----------
            unit: str
                unit in which to return the size
                can be bytes (default), mega or giga

            Returns
            -------
            size: float
        """
        size = 0

        for cachefile in glob.iglob('%s/*' % self._cache.cache):
            size += os.path.getsize(cachefile)

            if cachefile.endswith('.alt'):
                f = file(cachefile, "rb")
                _altpath = f.read()
                f.close()

                if os.path.exists(_altpath):
                    size += os.path.getsize(_altpath)

        return bytes_to_human(size, unit)

    def available_disk(self, path=None, unit='bytes'):
        """ Available disk on partition. Default location is cache folder.
        """
        if path is None:
            path = self._cache.cache

        if not os.path.isdir(path):
            path = os.path.dirname(path.rstrip('/'))

        if _platform == 'Windows':
            available = ctypes.c_ulonglong(0)

            ctypes.windll.kernel32.GetDiskFreeSpaceExW(
                ctypes.c_wchar_p(path),
                None,
                None,
                ctypes.pointer(available))

            return bytes_to_human(available.value, unit)
        else:
            cache_st = os.statvfs(path)

            return bytes_to_human(cache_st.f_bavail * cache_st.f_bsize, unit)

    def used_disk(self, path=None, unit='bytes'):
        """ Used disk on partition. Default location is cache folder.
        """
        if path is None:
            path = self._cache.cache

        return self.total_disk(path, unit) - self.available_disk(path, unit)

    def total_disk(self, path=None, unit='bytes'):
        """ Total disk on partition. Default location is cache folder.
        """
        if path is None:
            path = self._cache.cache

        if not os.path.isdir(path):
            path = os.path.dirname(path.rstrip('/'))

        if _platform == 'Windows':
            total = ctypes.c_ulonglong(0)

            ctypes.windll.kernel32.GetDiskFreeSpaceExW(
                ctypes.c_wchar_p(path),
                None,
                ctypes.pointer(total),
                None,
                )

            return bytes_to_human(total.value, unit)
        else:
            cache_st = os.statvfs(path)

            return bytes_to_human(cache_st.f_blocks * cache_st.f_bsize, unit)

    def disk_ready(self, path=None, ready_ratio=90.0):
        """ Checks the status of the disk. Basically if there is enough space
            left. The default location is cache dir.
        """
        if path is None:
            path = self._cache.cache

        disk_ratio = ( float(self.used_disk(path)) / \
                           float(self.total_disk(path)) * 100
                       )

        return disk_ratio < ready_ratio, disk_ratio

    def set_usage(self, mode=None, expiration=1.0):
        """ Customize cache usage.

            Parameters
            ----------
            mode: string
                'online' or 'offline':
                    - online will always query the server to have up
                      to date data.
                    - offline will only try to query the server if the
                      data is not cached.
            expiration: float
                Relevant only to online mode. The cache has an
                expiration mechanism. If two queries on the same
                resource are issued under the specified value, the
                cache will be used and the server will not be
                requested.
        """
        if mode == 'online':
            self._intf._mode = 'online'
            self._intf._memtimeout = expiration
        elif mode == 'offline':
            self._intf._mode = 'offline'
            self._intf._memtimeout = expiration
        else:
            self._intf._memtimeout = expiration



import os

from .. import Interface

_modulepath = os.path.dirname(os.path.abspath(__file__))

central = Interface('https://central.xnat.org', 'nosetests', 'nosetests')

def test_global_scan_listing():
    assert central.array.scans(project_id='CENTRAL_OASIS_CS', 
                               experiment_type='xnat:mrSessionData', 
                               scan_type='xnat:mrScanData'
                               )




import os
import tempfile

from .. import jsonutil

_csv_example = os.path.join(os.path.dirname(os.path.abspath(__file__)), 
                            'results.csv')

_list_of_dirs = jsonutil.csv_to_json(open(_csv_example, 'rb').read())

jtable = jsonutil.JsonTable(_list_of_dirs)

# coverage_tests
def test_dump_methods():
    jtable.dumps_csv()
    jtable.dumps_json()

    jtable.dump_csv(tempfile.mkstemp()[1])
    jtable.dump_json(tempfile.mkstemp()[1])

    jtable.where(psytool_audit_parentdata_audit9='0', 
                 psytool_audit_parentdata_audit2='0', 
                 psytool_audit_parentdata_audit1='2', 
                 psytool_audit_parentdata_audit10='2')

    jtable.where('5154')

    assert True

# unit_tests
def test_get_item_str_value():
    assert jtable['subject_label'] == jtable.get('subject_label')

def test_get_item_int_value():
    assert jtable.data[0] == jtable[0].data[0]

def test_get_item_list_value():
    assert set(jtable[['projects', 'subjectid']].headers()) == set(['projects', 'subjectid'])

def test_join():
    index = jtable.headers().index('subjectid')
    a = jtable.select(jtable.headers()[:index+1])
    b = jtable.select(jtable.headers()[index:])
    c = a.join('subjectid', b)
    assert len(jtable.headers()) == len(c.headers())







from .. import Interface

central = Interface('https://central.xnat.org', 'nosetests', 'nosetests')

def test_xpath_checkout():
    central.xpath.checkout(subjects=['OAS1_0001', 'OAS1_0002'])
    assert 'OAS1_0001' in central.xpath.subjects() and \
        'OAS1_0002' in central.xpath.subjects()

def test_elements():
    assert 'fs:region' in central.xpath.elements()

def test_keys():
    assert 'ID' in central.xpath.keys()

def test_values():
    assert 'OAS1_0002' in central.xpath.values('ID')

def test_element_attrs():
    assert isinstance(central.xpath.element_attrs('fs:region'), list)

    assert set(['SegId', 'hemisphere', 'name']).issubset(
        central.xpath.element_keys('fs:region'))

    assert 'Left-Putamen' in \
        central.xpath.element_values('fs:region', 'name')




#!/usr/bin/env python
"""Wrapper to run setup.py using setuptools."""

import zipfile
import os
import sys

from setuptools import Command
from sphinx_pypi_upload import UploadDoc

################################################################################
# Code to copy the sphinx-generated html docs in the distribution.
DOC_BUILD_DIR = os.path.join('build', 'sphinx', 'html')

def relative_path(filename):
    """ Return the relative path to the file, assuming the file is
        in the DOC_BUILD_DIR directory.
    """
    length = len(os.path.abspath(DOC_BUILD_DIR)) + 1
    return os.path.abspath(filename)[length:]


class ZipHelp(Command):
    description = \
    """zip the help created by the build_sphinx, and put it in the source distribution. """

    user_options = [
        ('None', None, 'this command has no options'),
        ]
 
    def run(self):
        if not os.path.exists(DOC_BUILD_DIR):
            raise OSError('Doc directory does not exist.')
        target_file = os.path.join('doc', 'documentation.zip')
        # ZIP_DEFLATED actually compresses the archive. However, there
        # will be a RuntimeError if zlib is not installed, so we check
        # for it. ZIP_STORED produces an uncompressed zip, but does not
        # require zlib.
        try:
            zf = zipfile.ZipFile(target_file, 'w', 
                                            compression=zipfile.ZIP_DEFLATED)
        except RuntimeError:
            zf = zipfile.ZipFile(target_file, 'w', 
                                            compression=zipfile.ZIP_STORED)    

        for root, dirs, files in os.walk(DOC_BUILD_DIR):
            relative = relative_path(root)
            if not relative.startswith('.doctrees'):
                for f in files:
                    zf.write(os.path.join(root, f), 
                            os.path.join(relative, f))
        zf.close()

    def initialize_options(self):
        pass
    
    def finalize_options(self):
        pass


class GenerateHelp(Command):
    description = " Generate the autosummary files "

    user_options = [
        ('None', None, 'this command has no options'),
        ]
 
    def run(self):
        os.system('%s doc/sphinxext/autosummary_generate.py -o doc/generated/ doc/*.rst'
                  % sys.executable)


    def initialize_options(self):
        pass
    
    def finalize_options(self):
        pass

    

################################################################################
# Call the setup.py script, injecting the setuptools-specific arguments.

extra_setuptools_args = dict(
                            tests_require=['nose', 'coverage'],
                            test_suite='nose.collector',
                            cmdclass={'zip_help':ZipHelp,
                                      'generate_help':GenerateHelp,
                                      'upload_help': UploadDoc},
                            zip_safe=False,
                            )


if __name__ == '__main__':
    execfile('setup.py', dict(__name__='__main__', 
                          extra_setuptools_args=extra_setuptools_args))





from .. import uriutil

def test_translate_uri():
    assert uriutil.translate_uri('/assessors/out_resources/files') == '/assessors/out/resources/files'
    assert uriutil.translate_uri('/assessors/out_resource/files') == '/assessors/out/resource/files'
    assert uriutil.translate_uri('/assessors/in_resources/files') == '/assessors/in/resources/files'
    assert uriutil.translate_uri('/assessors/in_resource/files') == '/assessors/in/resource/files'

def test_inv_translate_uri():
    assert uriutil.inv_translate_uri('/assessors/out/resources/files') == '/assessors/out_resources/files'
    assert uriutil.inv_translate_uri('/assessors/out/resource/files') == '/assessors/out_resource/files'
    assert uriutil.inv_translate_uri('/assessors/in/resources/files') == '/assessors/in_resources/files'
    assert uriutil.inv_translate_uri('/assessors/in/resource/files') == '/assessors/in_resource/files'

def test_join_uri():
    assert uriutil.join_uri('/projects', 'project_id', 'subjects', 'subject_id') == '/projects/project_id/subjects/subject_id'

def test_uri_last():
    assert uriutil.uri_last('/projects/1/subjects/2') == '2'

def test_uri_nextlast():
    assert uriutil.uri_nextlast('/projects/1/subjects/2') == 'subjects'

def test_uri_parent():
    assert uriutil.uri_parent('/projects/1/subjects/2') == '/projects/1/subjects'

def test_uri_grandparent():
    assert uriutil.uri_grandparent('/projects/1/subjects/2') == '/projects/1'

def test_uri_split():
    assert uriutil.uri_split('/projects/1/subjects/2') == ['/projects/1/subjects', '2']




import time
import platform
import socket

from lxml import etree
from lxml.etree import Element, QName

from .uriutil import uri_parent
from .jsonutil import JsonTable
from . import httputil

_nsmap = {'xnat':'http://nrg.wustl.edu/xnat',
          'prov':'http://www.nbirn.net/prov',
          'xsi':'http://www.w3.org/2001/XMLSchema-instance'
          }

_required = ['program', 'timestamp', 'user', 'machine', 'platform']
_optional = ['program_version', 'program_arguments', 
             'cvs', 
             'platform_version', 
             'compiler', 'compiler_version', 
             'library', 'library_version'
             ]

_all = ['program', 'program_version', 'program_arguments',
        'timestamp',
        'cvs', 
        'user',
        'machine',
        'platform', 'platform_version', 
        'compiler', 'compiler_version', 
        # 'library', 'library_version'
        ]

_platform_name, _hostname, \
_platform_version, _platform_version2,\
_machine, _machine2 = platform.uname()
_machine = socket.gethostname()


def provenance_document(eobj, process_steps, overwrite):
    root_node = etree.fromstring(eobj.get())

    existing_prov = None
    for child in root_node.getchildren():
        if str(child.tag).endswith('provenance'):
            existing_prov = child
            break

    if existing_prov is not None and not overwrite:
        prov_node = existing_prov
    else:
        if existing_prov is not None and overwrite:
            root_node.remove(existing_prov)

        prov_node = Element(QName(_nsmap['xnat'], 'provenance'), 
                            nsmap=_nsmap
                            )
        root_node.insert(0, prov_node)

    prov_node.extend(provenance_parameters(process_steps))

    return etree.tostring(root_node.getroottree())


def provenance_parameters(process_steps):
    prov = []

    for step in process_steps:

        if not set(_required).issubset(step.keys()):
            missing = list(set(_required).difference(step.keys()))

            raise Exception(('Following attributes are '
                             'required to define provenance: %s' % missing
                             )
                            )

        prov.append(process_step_xml(**step))

    return prov

def process_step_xml(**kwargs):
    
    step_node = Element(QName(_nsmap['prov'], 'processStep'), nsmap=_nsmap)

    program_node = Element(QName(_nsmap['prov'], 'program'), nsmap=_nsmap)
    program_node.text = kwargs['program']

    if kwargs.has_key('program_version'):
        program_node.set('version', kwargs['program_version'])

    if kwargs.has_key('program_arguments'):
        program_node.set('arguments', kwargs['program_arguments'])

    step_node.append(program_node)

    timestamp_node = Element(QName(_nsmap['prov'], 'timestamp'), 
                             nsmap=_nsmap
                             )
    timestamp_node.text = kwargs['timestamp']

    step_node.append(timestamp_node)

    if kwargs.has_key('cvs'):
        cvs_node = Element(QName(_nsmap['prov'], 'cvs'), nsmap=_nsmap)
        cvs_node.text = kwargs['cvs']
        
        step_node.append(cvs_node)

    user_node = Element(QName(_nsmap['prov'], 'user'), nsmap=_nsmap)
    user_node.text = kwargs['user']

    step_node.append(user_node)

    machine_node = Element(QName(_nsmap['prov'], 'machine'), nsmap=_nsmap)
    machine_node.text = kwargs['machine']

    step_node.append(machine_node)

    platform_node = Element(QName(_nsmap['prov'], 'platform'), nsmap=_nsmap)
    platform_node.text = kwargs['platform']

    if kwargs.has_key('platform_version'):
        platform_node.set('version', kwargs['platform_version'])

    step_node.append(platform_node)

    if kwargs.has_key('compiler'):
        compiler_node = Element(QName(_nsmap['prov'], 'compiler'), 
                                nsmap=_nsmap
                                )
        compiler_node.text = kwargs['compiler']

        if kwargs.has_key('compiler_version'):
            compiler_node.set('version', kwargs['compiler_version'])

        step_node.append(compiler_node)

    if kwargs.has_key('library'):
        library_node = Element(QName(_nsmap['prov'], 'library'), 
                                nsmap=_nsmap
                                )
        library_node.text = kwargs['library']

        if kwargs.has_key('library_version'):
            library_node.set('version', kwargs['library_version'])

        step_node.append(library_node)

    return step_node


class Provenance(object):
    """ Class to annotate processed data with provenance information.
        The following parameters are available:
            - program
            - program_version
            - program_arguments
            - timestamp
            - cvs
            - user
            - machine
            - platform
            - platform_version
            - compiler
            - compiler_version
        
        Examples
        --------
            >>> prov = {'program':'young',
                        'timestamp':'2011-03-01T12:01:01.897987', 
                        'user':'angus', 
                        'machine':'war', 
                        'platform':'linux',
                        }
            >>> element.provenance.set(prov)
            >>> element.provenance.get()
            >>> element.delete()
    """

    def __init__(self, eobject):
        self._intf = eobject._intf
        self._eobject = eobject

    def set(self, process_steps, overwrite=False):
        """ Set provenance information for the data within this element.

            .. note::

                If some required parameters are not provided, theses
                parameters will be extracted from the current machine
                and set automatically. Those parameters are:
                    - machine
                    - platform
                    - timestamp
                    - user

            .. warning::
                overwrite option doesn't work because of a bug with the 
                allowDataDeletion flag in XNAT

            Parameters
            ----------
            process_steps: list or dict
                dict or list of dicts to define the processing steps
                of the data. The minimum set of information to give
                is: program, timestamp, user, machine and platform. More
                keywords in the class documentation.
            overwrite: boolean
                If False the process_steps are added to the existing ones.
                Else the processing steps overwrite any existing provenance.
        """
        if isinstance(process_steps, dict):
            process_steps = [process_steps]

        for process_step in process_steps:
            _timestamp = time.strftime('%Y-%m-%dT%H:%M:%S', 
                                       time.localtime()
                                       )
            
            if not process_step.has_key('machine'):
                process_step['machine'] = _machine
            if not process_step.has_key('platform'):
                process_step['platform'] = _platform_name
                process_step['platform_version'] = _platform_version
            if not process_step.has_key('timestamp'):
                process_step['timestamp'] = _timestamp
            if not process_step.has_key('user'):
                process_step['user'] = self._intf._user

        doc = provenance_document(self._eobject, process_steps, overwrite)

        body, content_type = httputil.file_message(
            doc, 'text/xml', 'prov.xml', 'prov.xml')

        prov_uri = self._eobject._uri

        if overwrite:
            prov_uri += '?allowDataDeletion=true'

        self._intf._exec(prov_uri, 
                         method='PUT', 
                         body=body,
                         headers={'content-type':content_type}
                         )

    def get(self):
        """ Gets all the provenance information for that object.

            Returns
            -------
            A list of dicts.
        """
        datatype = self._eobject.datatype()

        columns = ['%s/ID' % datatype] + [
            '%s/provenance/processStep/%s' % (datatype, field)
            for field in _all
            ]       

        prov_uri = uri_parent(self._eobject._uri)
        prov_uri += '?columns='
        prov_uri += ','.join(columns)

        steps = []

        table = JsonTable(self._intf._get_json(prov_uri))

        id_header = 'ID' if table.has_header('ID') \
            else '%s/id' % datatype.lower()

        for step in table.where(**{id_header:self._eobject.id()}):
            step_dict = {}
            for key in step.keys():
                if 'processstep' in key:
                    step_dict[key.split('processstep/')[1]] = step[key]

            steps.append(step_dict)

        return steps

    def delete(self):
        """ Removes the provenance attached to this object.

            .. warning::
                doesn't work because of a bug with the allowDataDeletion
                flag in XNAT
        """

        provenance_node = self._eobject.xpath('//xnat:provenance')

        if provenance_node != []:
            provenance_node = provenance_node[0]

            parent_node = provenance_node.getparent()
            parent_node.remove(provenance_node)

            doc = etree.tostring(parent_node.getroottree())

            body, content_type = httputil.file_message(
                doc, 'text/xml', 'prov.xml', 'prov.xml')

            self._intf._exec(
                '%s?allowDataDeletion=true' % self._eobject._uri, 
                method='PUT', 
                body=body,
                headers={'content-type':content_type}
                )
            



import re
from datetime import datetime
from os import path
from glob import iglob

from lxml import etree

from .jsonutil import JsonTable


def get_subject_id(location):
    f = open(location, 'rb')
    content = f.read()
    f.close()

    subject = re.findall('<xnat:Subject\sID="(.*?)"\s', content)

    if subject != []:
        return subject[0]


class XpathStore(object):

    """ This utility class is used to query XML files describing XNAT 
        subjects but which are stored in the cache directory. Xpath is
        used as the query lanaguage but a few helpers are provided
        to make simple queries.
    """
    
    def __init__(self, interface):
        self._intf = interface
        self._tree = None
        self._nsmap = {}

    def __call__(self, xpath):
        if self._tree is None:
            self._load()

        return self._tree.xpath(xpath, namespaces=self._nsmap)

    def _load(self):
        roots = []
        nsmap = {}

        for location in iglob('%s/*.xml' % self._intf._cachedir):
            f = open(location, 'rb')
            content = f.read()
            f.close()

            if get_subject_id is not None:
                roots.append(etree.fromstring(content))
                nsmap.update(roots[-1].nsmap)

        self._tree = etree.Element('Store')
        self._tree.extend(roots)
        self._nsmap = nsmap

    def _last_modified(self):
        entry_point = self._intf._get_entry_point()
        uri = '%s/subjects?columns=last_modified' % entry_point

        return dict(JsonTable(self._intf._get_json(uri), 
                              order_by=['ID', 'last_modified']
                              ).select(['ID', 'last_modified']
                                       ).items()
                    )

    def update(self):
        """ Updates the xml files in the cachedir.
        """
        last_modified = self._last_modified()
        
        for location in iglob('%s/*.xml' % self._intf._cachedir):

            subject_id = get_subject_id(location)

            if subject_id is not None:
                local_time = path.getmtime(location)
                local_time = datetime.fromtimestamp(local_time)

                server_time = last_modified[subject_id].split('.')[0]
                server_time = datetime.strptime(server_time, 
                                                '%Y-%m-%d %H:%M:%S')

                if local_time < server_time:
                    self._exec('%s/subjects/%s' % (
                            self._intf._get_entry_point(), subject_id))

    def checkout(self, project=None, subjects=None):
        """ Downloads all the subject XMLs for a project or a list
            of subjects.

            Parameters
            ----------
            project: str
                The id of the project
            subjects: list
                List of subjects that have to be downloaded.
        """
        if project is not None and subjects is None:
            for s in self._intf.select('/project/%s/subjects' % project):
                s.get()
        elif subjects is not None:
            for sid in subjects:
                self._intf._exec('%s/subjects/%s?format=xml' % (
                        self._intf._get_entry_point(), sid))

        self._load()

                    
    def subjects(self):
        """ Returns all the subject ids.
        """
        return self.__call__('//xnat:Subject/attribute::ID')

    def keys(self):
        """ Returns attribute keys at the subject level.
        """
        return self.element_keys('xnat:Subject')

    def values(self, key):
        """ Returns the values for all subjects for a specific attribute.
        """
        return self.element_values('xnat:Subject', key)

    def attrs(self):
        """ Returns the attributes of all subjects.
        """
        return self.element_attrs('xnat:Subject')

    def elements(self):
        """ Returns the element names for all subjects elements.

            .. note:: in xpath terms, this function is returning all
                the names of the subjects descendant nodes.
        """
        elements = set()

        for element in self.__call__('//xnat:Subject/descendant::*'):
            for ns in element.nsmap:
                if element.nsmap[ns] in element.tag:
                    n = element.tag.replace('{%s}' % element.nsmap[ns], 
                                            '%s:' % ns)
                    elements.add(n)

        return list(elements)

    def element_attrs(self, name):
        """ Returns the attributes of this specific element.
        """
        attrs = []

        for element in self.__call__('//%s' % name):
            attrs.append(element.attrib)
        
        return attrs

    def element_keys(self, name):
        """ Returns the attribute keys of this specific element.
        """
        keys = set()

        for element in self.__call__('//%s' % name):
            for element_key in element.keys():
                keys.add(element_key)
        
        return list(keys)

    def element_values(self, name, key):
        """ Returns the attribute values of this specific element.
        """
        values = set()

        for subject in self.__call__('//%s' % name):
            values.add(subject.get(key))
        
        return list(values)

    def element_text(self, name):
        """ Returns the text values of this specific element.
        """
        return list(set(self.__call__('//%s/child::text()' % name)))




import os
from functools import partial

# default path to xnat pass file
def path():
    return os.getenv('USERPROFILE') or os.getenv('HOME') + "/.xnatPass"

# str ->  {'host': ..., 'u': ..., 'p': ..., 'port': ...} | None
def read_xnat_pass(f):
    if os.path.exists(f) and os.path.isfile(f):
        infile = open(f)
        return parse_xnat_pass(infile.readlines())
    else:
        # raise IOError('XNAT Pass file :' + f + " does not exist")
        return None

# [str] -> {'host': ..., 'u': ..., 'p': ..., 'port':...} | None
def parse_xnat_pass(lines):
    empty = {'host': None,  'u': None, 'p': None}
    line = find_plus_line(lines)
    u = ('u',partial(find_token,'@'),True)
    host = ('host',partial(find_token,'='),True)
    p = ('p',partial(lambda x : (x,x)),True)
    
    def update_state(x,k,state):
        state[k] = x
        return state
    if line == None:
        return None
    else:
        return chain([u,host,p],line,empty,update_state)

# [(str, str -> str, bool)] ->
#  str ->
#  dict
#  dict -> dict
#  dict | None
def chain(ops, initEnv, initState, update_statef):
    env = initEnv
    state = initState
    for op in ops:
        (k,_op,mandatory) = op
        tmp = _op(env)
        if tmp == None:
            if mandatory:
                return None
        else:
            (v,rest) = tmp
            env = rest
            state = update_statef(v,k,state)
    return state
        
# [str] -> str | None
def find_plus_line(lines):
    plusLines = filter (lambda x: x.startswith('+'), lines)
    if len(plusLines) == 0:
        return None
    else:
        return plusLines[0][1:]

# char -> str -> (str,str) | None
def find_token(tok,line):
    splitString = map(lambda x: x.strip(), line.split(tok))
    if len(splitString) == 0 or len(splitString) == 1 or splitString[0] == '':
        return None
    else:
        return (splitString[0],splitString[1])




import os
import re
import time
import tempfile
import email
import getpass

import httplib2
import json

try:
    import socks
except ImportError:
    socks = None
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse
from .select import Select
from .cache import CacheManager, HTCache
from .help import Inspector, GraphData, PaintGraph, _DRAW_GRAPHS
from .manage import GlobalManager
from .uriutil import join_uri, file_path, uri_last
from .jsonutil import csv_to_json
from .errors import is_xnat_error
from .errors import catch_error
from .array import ArrayData
from .xpath_store import XpathStore
from . import xpass


DEBUG = False


# main entry point
class Interface(object):
    """ Main entry point to access a XNAT server.

        >>> central = Interface(server='http://central.xnat.org:8080',
                                user='login',
                                password='pwd',
                                cachedir='/tmp')

        Or with config file:

        >>> central = Interface(config='/home/me/.xnat.cfg')

        Or for interactive use:

        >>> central = Interface('http://central.xnat.org')

        .. note::
            The interactive mode is activated whenever an argument within
            server, user or password is missing. In interactive mode pyxnat
            tries to check the validity of the connection parameters.

        Or anonymously (unauthenticated):

        >>> central = Interface('http://central.xnat.org', anonymous=True)

        Attributes
        ----------
        _mode: online | offline
            Online or offline mode
        _memtimeout: float
            Lifespan of in-memory cache

        .. note::
            Proxy support requires the socks module be installed. This can be
            installed via pip:
            pip install SocksiPy-branch
    """

    def __init__(self, server=None, user=None, password=None,
                 cachedir=tempfile.gettempdir(), config=None,
                 anonymous=False, proxy=None):
        """
            Parameters
            ----------
            server: string | None
                The server full URL (including port and XNAT instance name
                if necessary) e.g. http://central.xnat.org,
                http://localhost:8080/xnat_db
                Or a path to an existing config file. In that case the other
                parameters (user etc..) are ignored if given.
                If None the user will be prompted for it.
            user: string | None
                A valid login registered through the XNAT web interface.
                If None the user will be prompted for it.
            password: string | None
                The user's password.
                If None the user will be prompted for it.
            cachedir: string
                Path of the cache directory (for all queries and
                downloaded files)
                If no path is provided, a platform dependent temp dir is
                used.
            config: string
                Reads a config file in json to get the connection parameters.
                If a config file is specified, it will be used regardless of
                other parameters that might have been given.
            anonymous: boolean
                Indicates an unauthenticated connection.  If True, user
                and password are ignored and a session is started with
                no credentials.
            proxy: string | None
                Indicates the full URL for an HTTP proxy server to be used for
                transactions with the specified XNAT server. If you need to
                specify a username and password for proxy access, prepend them
                to the hostname:
                http://user:pass@hostname:port

        """

        self._interactive = False

        self._anonymous = anonymous

        if self._anonymous:

            if server is None:
                self._server = raw_input('Server: ')
                self._interactive = True
            else:
                self._server = server
                self._interactive = False

            self.__set_proxy(proxy)

            self._user = None
            self._pwd = None

            self._cachedir = os.path.join(
                cachedir, 'anonymous@%s' % self._server.split('//')[1].replace(
                    '/', '.').replace(':', '_')
            )

        else:

            if not all([server, user, password]) and not config:
                self._interactive = True

            if all(arg is None
                    for arg in [server, user, password, config]) \
                    and os.path.exists(xpass.path()):

                connection_args = xpass.read_xnat_pass(xpass.path())

                if connection_args is None:
                    raise Exception('XNAT configuration file not found '
                                    'or formatted incorrectly.')

                self._server = connection_args['host']
                self._user = connection_args['u']
                self._pwd = connection_args['p']
                self._cachedir = os.path.join(
                    cachedir, '%s@%s' % (
                        self._user,
                        self._server.split('//')[1].replace(
                            '/', '.').replace(':', '_')
                    )
                )

                if 'proxy' in connection_args:
                    self.__set_proxy(connection_args['proxy'])
                else:
                    self.__set_proxy(None)

            elif config is not None:
                self.load_config(config)

            else:
                if server is None:
                    self._server = raw_input('Server: ')
                else:
                    self._server = server

                if user is None:
                    user = raw_input('User: ')

                if password is None:
                    password = getpass.getpass()

                self._user = user
                self._pwd = password

                self._cachedir = os.path.join(
                    cachedir, '%s@%s' % (
                        self._user,
                        self._server.split('//')[1].replace(
                            '/', '.').replace(':', '_')
                    )
                )

                self.__set_proxy(proxy)

        self._callback = None

        self._memcache = {}
        self._memtimeout = 1.0
        self._mode = 'online'
        self._struct = {}
        self._entry = None

        self._last_memtimeout = 1.0
        self._last_mode = 'online'

        self._jsession = 'authentication_by_credentials'
        self._connect_extras = {}
        self._connect()

        self.inspect = Inspector(self)
        self.select = Select(self)
        self.array = ArrayData(self)
        self.cache = CacheManager(self)
        self.manage = GlobalManager(self)
        self.xpath = XpathStore(self)

        if _DRAW_GRAPHS:
            self._get_graph = GraphData(self)
            self.draw = PaintGraph(self)

        if self._anonymous:
            response, content = self._http.request(self._server, 'GET')
            self._jsession = response['set-cookie'][:44]

        if self._interactive:
            self._get_entry_point()

        self.inspect()

    def __getstate__(self):
        return {
            '_server': self._server,
            '_user': self._user,
            '_pwd': self._pwd,
            '_cachedir': os.path.split(self._cachedir)[0],
            '_anonymous': self._anonymous,
        }

    def __setstate__(self, dictionary):
        self.__dict__ = dictionary
        if self._anonymous:
            self.__init__(self._server, anonymous=True)
        else:
            self.__init__(self._server, self._user, self._pwd, self._cachedir)

    def __set_proxy(self, proxy=None):
        if proxy is None:
            proxy = os.environ.get("http_proxy")
        if proxy is None:
            self._proxy_url = None
        else:
            self._proxy_url = urlparse(proxy)

    def _get_entry_point(self):
        if self._entry is None:
            # /REST for XNAT 1.4, /data if >=1.5
            self._entry = '/REST'
            try:
                self._jsession = 'JSESSIONID=' + self._exec('/data/JSESSION')
                self._entry = '/data'

                if is_xnat_error(self._jsession):
                    catch_error(self._jsession)
            except Exception as e:
                if not '/data/JSESSION' in str(e):
                    raise e

        return self._entry

    def _connect(self, **kwargs):
        """ Sets up the connection with the XNAT server.

            Parameters
            ----------
            kwargs: dict
                Can be used to pass additional arguments to
                the Http constructor. See the httplib2 documentation
                for details. http://code.google.com/p/httplib2/
        """

        if kwargs != {}:
            self._connect_extras = kwargs
        else:
            kwargs = self._connect_extras

        kwargs['disable_ssl_certificate_validation'] = True

        # If a proxy was configured, then add that in.
        if socks and self._proxy_url is not None:
            if self._proxy_url.username is None:
                kwargs['proxy_info'] = httplib2.ProxyInfo(
                    socks.PROXY_TYPE_HTTP,
                    self._proxy_url.hostname,
                    self._proxy_url.port)
            else:
                kwargs['proxy_info'] = httplib2.ProxyInfo(
                    socks.PROXY_TYPE_HTTP,
                    self._proxy_url.hostname,
                    self._proxy_url.port,
                    None,
                    self._proxy_url.username,
                    self._proxy_url.password)

        if DEBUG:
            httplib2.debuglevel = 2

        # compatibility with httplib2 < 0.7
        try:
            self._http = httplib2.Http(HTCache(self._cachedir, self),
                                       **kwargs)
        except:
            del kwargs['disable_ssl_certificate_validation']
            self._http = httplib2.Http(
                HTCache(self._cachedir, self),
                **kwargs
            )

        if not self._anonymous:
            self._http.add_credentials(self._user, self._pwd)

    def _exec(self, uri, method='GET', body=None, headers=None):
        """ A wrapper around a simple httplib2.request call that:
                - avoids repeating the server url in the request
                - deals with custom caching mechanisms
                - manages a user session with cookies
                - catches and broadcast specific XNAT errors

            Parameters
            ----------
            uri: string
                URI of the resource to be accessed. e.g. /REST/projects
            method: GET | PUT | POST | DELETE
                HTTP method.
            body: string
                HTTP message body
            headers: dict
                Additional headers for the HTTP request.
        """

        if headers is None:
            headers = {}

        self._get_entry_point()

        uri = join_uri(self._server, uri)

        if DEBUG:
            print(uri)
        # using session authentication
        headers['cookie'] = self._jsession
        headers['connection'] = 'keep-alive'

        # reset the memcache when client changes something on the server
        if method in ['PUT', 'DELETE']:
            self._memcache = {}

        # Initialize these to default values.
        response = None
        info = None
        content = None

        if self._mode == 'online' and method == 'GET':

            if time.time() - self._memcache.get(uri, 0) < self._memtimeout:
                if DEBUG:
                    print('send: GET CACHE %s' % uri)

                info, content = self._http.cache.get(uri
                                                     ).split('\r\n\r\n', 1)

                self._memcache[uri] = time.time()
            else:
                response, content = self._http.request(uri, method,
                                                       body, headers)
                self._memcache[uri] = time.time()

        elif self._mode == 'offline' and method == 'GET':

            cached_value = self._http.cache.get(uri)

            if cached_value is not None:
                if DEBUG:
                    print('send: GET CACHE %s' % uri)
                info, content = cached_value.split('\r\n\r\n', 1)
            else:
                try:
                    self._http.timeout = 10

                    response, content = self._http.request(uri, method,
                                                           body, headers)

                    self._http.timeout = None
                    self._memcache[uri] = time.time()
                except Exception as e:
                    catch_error(e)
        else:
            response, content = self._http.request(uri, method,
                                                   body, headers)

        if DEBUG:
            if response is None:
                response = httplib2.Response(email.message_from_string(info))
                print('reply: %s %s from cache') % (response.status,
                                                   response.reason
                                                   )
                for key in response.keys():
                    print('header: %s: %s') % (key.title(), response.get(key))

        if response is not None and 'set-cookie' in response:
            cookies = response.get('set-cookie')
            jsessionid = re.findall(r'(JSESSIONID=[0-9A-F]+);', cookies)
            if len(jsessionid) > 0:
                self._jsession = jsessionid[0]

        if (response is not None and response.status == 404) or is_xnat_error(content):
            if DEBUG:
                print(response.keys())
                print(response.get("status"))

            catch_error(content)

        return content

    def _get_json(self, uri):
        """ Specific Interface._exec method to retrieve data.
            It forces the data format to csv and then puts it back to a
            json-like format.

            Parameters
            ----------
            uri: string
                URI of the resource to be accessed. e.g. /REST/projects

            Returns
            -------
            List of dicts containing the results
        """
        if 'format=json' in uri:
            uri = uri.replace('format=json', 'format=csv')
        else:
            if '?' in uri:
                uri += '&format=csv'
            else:
                uri += '?format=csv'

        content = self._exec(uri, 'GET')

        if is_xnat_error(content):
            catch_error(content)

        json_content = csv_to_json(content)

        # add the (relative) path field for files
        base_uri = uri.split('?')[0]
        if uri_last(base_uri) == 'files':
            for element in json_content:
                element['path'] = file_path(element['URI'])
        return json_content

    def _get_head(self, uri):
        if DEBUG:
            print('GET HEAD')
        _nocache = httplib2.Http(**self._connect_extras)
        if self._user:
            _nocache.add_credentials(self._user, self._pwd)

        rheaders = {'cookie': self._jsession}

        try:
            head = _nocache.request(
                '%s%s' % (self._server, uri), 'HEAD', headers=rheaders)[0]
        except:
            time.sleep(1)
            head = _nocache.request(
                '%s%s' % (self._server, uri), 'HEAD', headers=rheaders)[0]

        info = email.Message.Message()

        for key, value in head.iteritems():
            if key == 'content-disposition':
                info['content-location'] = '%s%s' % (self._server, uri)
            if key not in ['set-cookie']:
                info[key] = value

        return info

    def save_config(self, location):
        """ Saves current configuration - including password - in a file.

            .. warning::
                Since the password is saved as well, make sure the file
                is saved at a safe location with appropriate permissions.

            .. note::
                This method raises NotImplementedError for an anonymous
                interface.

            Parameters
            ----------
            location: string
                Destination config file.
        """
        if self._anonymous:
            raise NotImplementedError(
                'no save_config() for anonymous interfaces')

        if not os.path.exists(os.path.dirname(location)):
            os.makedirs(os.path.dirname(location))

        fp = open(location, 'w')
        config = {'server': self._server,
                  'user': self._user,
                  'password': self._pwd,
                  'cachedir': os.path.split(self._cachedir)[0],
                  }
        if self._proxy_url:
            config['proxy'] = self._proxy_url.geturl()

        json.dump(config, fp)
        fp.close()

    def load_config(self, location):
        """ Loads a configuration file and replaces current connection
            parameters.

            .. note::
                This method raises NotImplementedError for an anonymous
                interface.

            Parameters
            ----------
            location: string
                Configuration file path.
        """
        if self._anonymous:
            raise NotImplementedError(
                'no load_config() for anonymous interfaces')

        if os.path.exists(location):
            fp = open(location, 'rb')
            config = json.load(fp)
            fp.close()

            self._server = str(config['server'])
            self._user = str(config['user'])
            self._pwd = str(config['password'])
            self._cachedir = str(config['cachedir'])

            self._cachedir = os.path.join(
                self._cachedir, '%s@%s' % (
                    self._user,
                    self._server.split('//')[1].replace(
                        '/', '.').replace(':', '_')
                )
            )

            if 'proxy' in config:
                self.__set_proxy(str(config['proxy']))
            else:
                self.__set_proxy(None)

        else:
            raise Exception('Configuration file does not exist.')

    def version(self):
        return self._exec('/data/version')

    def set_logging(self, level=0):
        pass

    def disconnect(self):
        """
            Tell XNAT to disconnect this session
        """
        self._exec('/data/JSESSION', method='DELETE')
        pass



import os
from uuid import uuid1

from .. import Interface

_modulepath = os.path.dirname(os.path.abspath(__file__))

central = Interface('https://central.xnat.org', 'nosetests', 'nosetests')

sid = uuid1().hex
eid = uuid1().hex

subject = central.select.project('nosetests').subject(sid)
experiment = subject.experiment(eid)

def test_fancy_resource_create():
    field_data = {'experiment':'xnat:mrSessionData',
                  'ID':'TEST_%s' % eid,
                  'xnat:mrSessionData/age':'42',
                  'xnat:subjectData/investigator/lastname':'doe',
                  'xnat:subjectData/investigator/firstname':'john',
                  'xnat:subjectData/ID': 'TEST_%s' % sid,
                  }

    experiment.create(**field_data)

    assert subject.exists()
    assert experiment.exists()

    globals()['subject'] = experiment.parent()
    globals()['experiment'] = experiment

def test_attr_get():
    assert experiment.attrs.get('xnat:mrSessionData/age') == '42.0'

def test_attr_mget():
    fields = ['xnat:subjectData/investigator/firstname', 
              'xnat:subjectData/investigator/lastname'
              ]

    assert subject.attrs.mget(fields) == ['john', 'doe']

def test_attr_set():
    experiment.attrs.set('xnat:mrSessionData/age', '26')
    assert experiment.attrs.get('xnat:mrSessionData/age') == '26.0'

def test_attr_mset():
    field_data = {'xnat:subjectData/investigator/firstname':'angus',
                  'xnat:subjectData/investigator/lastname':'young',
                  }

    subject.attrs.mset(field_data)

    assert set(subject.attrs.mget(field_data.keys())) == \
        set(field_data.values())

def test_cleanup():
    subject.delete()
    assert not subject.exists()



from uuid import uuid1

from .. import Interface

central = Interface('https://central.xnat.org', 'nosetests', 'nosetests')
project = central.select('/project/nosetests')

prov = {
    'program':'young',
    'timestamp':'2011-03-01T12:01:01.897987', 
    'user':'angus', 
    'machine':'war', 
    'platform':'linux',
    }

sid = uuid1().hex
eid = uuid1().hex
aid = uuid1().hex

assessor = project.subject(sid).experiment(eid).assessor(
    aid).insert(use_label=True)


def test_provenance():
    assessor.provenance.set(prov)
    _prov = assessor.provenance.get()[0]

    assert prov['program'] == _prov['program']

# def test_del_provenance():
#     assessor.provenance.delete()
#     print assessor.provenance.get()
#     assert assessor.provenance.get()[0] == []

def test_provenance_cleanup():
    project.subject(sid).delete()
    assert not project.subject(sid).exists()



from uuid import uuid1

from .. import Interface

central = Interface('https://central.xnat.org', 'nosetests', 'nosetests')
project = central.select('/project/nosetests')

variables = {'Subjects' : {'newgroup' : {'foo' : 'string', 'bar' : 'int'}}}

sid = uuid1().hex
eid = uuid1().hex
cid = uuid1().hex

scan = project.subject(sid).experiment(eid).scan(cid).insert(use_label=True)

# def test_add_custom_variables():
#     project.add_custom_variables(variables)

# def test_get_custom_variables():
#     assert project.get_custom_variables() == variables

def test_set_param():
    
    scan.set_param('foo', 'foostring')
    scan.set_param('bar', '1')

    assert scan.params() == ['foo', 'bar']
    
def test_get_params():
    assert scan.get_params() == ['foostring', '1']

def test_params_cleanup():
    project.subject(sid).delete()
    assert not project.subject(sid).exists()



from .. import Interface

central = Interface('https://central.xnat.org', 'nosetests', 'nosetests')

notified = []

def notify(cobj, eobj):
    notified.append(eobj)

def test_register_callback():
    local_notified = []
    central.manage.register_callback(notify)
    for element in central.select('/projects'):
        local_notified.append(element)

    assert notified == local_notified

def test_unregister_callback():
    central.manage.unregister_callback()
    assert central._callback is None




import os
import sys

from .interfaces import Interface
from .search import SearchManager
from .cache import CacheManager
from .select import Select
from .help import Inspector
from .users import Users



from uuid import uuid1
from .. import Interface

central = Interface('https://central.xnat.org', 'nosetests', 'nosetests')
central_anon = Interface('https://central.xnat.org', anonymous=True)

def test_simple_object_listing():
    assert isinstance(central.select.projects().get(), list)

def test_simple_path_listing():
    assert isinstance(central.select('/projects').get(), list)

def test_nested_object_listing():
    assert isinstance(central.select.projects('*OASIS*').subjects().get(), list)

def test_nested_path_listing():
    assert isinstance(central.select('/projects/*OASIS*/subjects').get(), list)

# def test_nested_for_access():
#     stop = False
#     for subject in central.select('/projects/*OASIS*/subjects'):
#         for f in subject.experiments().scans().resources().files():
#             assert isinstance(f._uri, (str, unicode))
#             stop =True
#             break
#         if stop:
#             break

def test_seach_access():
    constraints = [('xnat:subjectData/PROJECT', '=', 'CENTRAL_OASIS_CS'), 'AND']

    for subject in central.select('//subjects').where(constraints):
        assert '/projects/CENTRAL_OASIS_CS' in subject._uri

def test_anonymous_access():
    projects = central_anon.select.projects().get()
    assert isinstance(projects, list)
    assert list



import os
import zipfile

from .schema import class_name
from . import uriutil
from . import schema


def unzip(fzip,
          dest_dir,
          check={'run':lambda z,d: True,'desc':""}):
    """
    Extracts the given zip file to the given directory, but only if all members of the
    archive pass the given check.

        Parameters
        ----------
        src: fzip
            zipfile
        dest_dir: string
            directory into which to extract the archive
        check: dict
            An dictionary that has the keys:
                 'run' : A function that takes a filename and parent directory and returns Bool. By default
                         this function always returns True.
                 'dest' : A string description of this test. By default this is empty.

        Returns a tuple of type (bool,[string]) where if the extraction ran successfully the first is true and the
        second is a list of files that were extracted, and if not the first is false and the second is the name
        of the failing member.
    """
    for member in fzip.namelist():
        if not check['run'](member,dest_dir):
            return (False,member)

    fzip.extractall(path=dest_dir)
    return (True, map (lambda f: os.path.join(dest_dir,f),fzip.namelist()))

def download (dest_dir, instance=None,type="ALL",name=None, extract=False, safe=False):
    """
    Download all the files at this level that match the given constraint as a zip archive. Should not be called directly
    but from a instance of class that supports bulk downloading eg. "Scans"

        Parameters
        ----------
        instance : 'object
             The instance that contains local values needed by this function eg. instance._cbase stores the URI.
        dest_dir : string
             directory into which to place the downloaded archive
        type: string
             a comma separated list of file types, eg. "T1,T2". Default is "ALL"
        name: string
             the name of the zip archive. Defaults to None. See below for the default naming scheme
        extract: bool
             If True, the files are left extracted in the parent directory. Default is False
        safe: bool
             If true, run safety checks on the files before extracting, eg. check that the file
             doesn't exist in the parent directory before overwriting it. Default is False

        Default Zip Name
        ----------------
        Given the project "p", subject "s" and experiment "e", and that the "Scans" (as opposed to "Assessors"
        or "Reconstructions") are being downloaded, and the scan types are constrained to "T1,T2",
        the name of the zip file defaults to:
          "p_s_e_scans_T1_T2.zip"

        Exceptions
        ----------
        A generic Exception will be raised if any of the following happen:
         - This function is called directly and not from an instance of a class that supports bulk downloading eg."Scans"
         - The destination directory is unspecified
        A LookupError is raised if there are no resources to download
        A ValueError is raised if any of the following happen:
         - The project, subject and experiment names could not be extracted from the URI
         - The type constraint "ALL" is used with other constraints. eg. "ALL,T1,T2"
         - The URI associated with this class contains wildcards eg. /projects/proj/subjects/*/experiments/scans
        An EnvironmentError is raised if any of the following happen:
         - If "safe" is true, and (a) a zip file with the same name exists in given destination directory or
           (b) extracting the archive overrides an existing file. In the second case the downloaded archive
           is left in the parent directory.

        Return
        ------
        A path to the zip archive if "extract" is False, and a list of extracted files if True.
    """

    if instance is None:
        raise Exception('This function should be called directly but from an instance of a class that supports bulk downloading, eg. "Scans"')
    if dest_dir is None:
        raise Exception('Destination directory is unspecified')

    # the URI must be fully qualified with a project,subject and experiment
    if '%2A' in instance._cbase:
        raise ValueError('URI contains wildcards :' + instance._cbase)

    # Check that there are resources at this level
    available = instance.get()
    if len(available) == 0:
        raise LookupError(
            'There are no %s to download' % class_name(instance).lower())

    pse = uriutil.extract_uri(instance._cbase)
    if pse is None:
        raise ValueError("Could not extract project, subject and experiment from the uri: " + instance._cbase)

    # Extract the desired scan types. Clean up whitespace and remove dupes
    types = {}
    for t in type.split(','):
        cleaned = t.strip()
        if cleaned != "":
            types[cleaned] = cleaned

    # Make sure the user hasn't asked us to download ALL the scans and then asked
    # for them to be constrained to a type.
    if len(types) > 1 and 'ALL' in types:
        raise ValueError('The \"ALL\" scan type constraint cannot be used with any other constraint')

    (p,s,e) = pse

    # Make the name of the zip file
    default_zip_name = lambda: '%s_%s_%s_%s_%s' % (
        p, s, e, class_name(instance).lower(), '_'.join(types.values()))

    zip_name = name if name != None else default_zip_name()
    zip_location = os.path.join(dest_dir, zip_name + '.zip')

    if safe:
        if os.path.exists(zip_location):
            raise EnvironmentError("Unable to download to " + zip_location + " because this file already exists.")

    # Download from the server
    instance._intf._http.cache.preset(zip_location)
    instance._intf._exec(uriutil.join_uri(
            instance._cbase,','.join(types.values())) + '/files?format=zip')

    # Extract the archive
    fzip = zipfile.ZipFile(zip_location,'r')
    if extract:
        check = {'run': lambda f, d: not os.path.exists(os.path.join(dest_dir,f)),
                 'desc': 'File does not exist in the parent directory'}
        safeUnzip = lambda: unzip(fzip, dest_dir, check) if safe else lambda:unzip(fzip,dest_dir)
        (unzipped, paths) = safeUnzip()()
        if not unzipped:
            fzip.close()
            raise EnvironmentError("Unable to extract " + zip_location + " because file " + paths + " failed the following test: " + check['desc'])
        else:
            return paths
    else:
        fzip.close()
        return zip_location




import re, inspect, textwrap, pydoc
from docscrape import NumpyDocString, FunctionDoc, ClassDoc

class SphinxDocString(NumpyDocString):
    # string conversion routines
    def _str_header(self, name, symbol='`'):
        return ['.. rubric:: ' + name, '']

    def _str_field_list(self, name):
        return [':' + name + ':']

    def _str_indent(self, doc, indent=4):
        out = []
        for line in doc:
            out += [' '*indent + line]
        return out

    def _str_signature(self):
        return ['']
        if self['Signature']:
            return ['``%s``' % self['Signature']] + ['']
        else:
            return ['']

    def _str_summary(self):
        return self['Summary'] + ['']

    def _str_extended_summary(self):
        return self['Extended Summary'] + ['']

    def _str_param_list(self, name):
        out = []
        if self[name]:
            out += self._str_field_list(name)
            out += ['']
            for param,param_type,desc in self[name]:
                out += self._str_indent(['**%s** : %s' % (param.strip(),
                                                          param_type)])
                out += ['']
                out += self._str_indent(desc,8)
                out += ['']
        return out

    def _str_section(self, name):
        out = []
        if self[name]:
            out += self._str_header(name)
            out += ['']
            content = textwrap.dedent("\n".join(self[name])).split("\n")
            out += content
            out += ['']
        return out

    def _str_see_also(self, func_role):
        out = []
        if self['See Also']:
            see_also = super(SphinxDocString, self)._str_see_also(func_role)
            out = ['.. seealso::', '']
            out += self._str_indent(see_also[2:])
        return out

    def _str_warnings(self):
        out = []
        if self['Warnings']:
            out = ['.. warning::', '']
            out += self._str_indent(self['Warnings'])
        return out

    def _str_index(self):
        idx = self['index']
        out = []
        if len(idx) == 0:
            return out

        out += ['.. index:: %s' % idx.get('default','')]
        for section, references in idx.iteritems():
            if section == 'default':
                continue
            elif section == 'refguide':
                out += ['   single: %s' % (', '.join(references))]
            else:
                out += ['   %s: %s' % (section, ','.join(references))]
        return out

    def _str_references(self):
        out = []
        if self['References']:
            out += self._str_header('References')
            if isinstance(self['References'], str):
                self['References'] = [self['References']]
            out.extend(self['References'])
            out += ['']
        return out

    def __str__(self, indent=0, func_role="obj"):
        out = []
        out += self._str_signature()
        out += self._str_index() + ['']
        out += self._str_summary()
        out += self._str_extended_summary()
        for param_list in ('Parameters', 'Attributes', 'Methods',
                           'Returns','Raises'):
            out += self._str_param_list(param_list)
        out += self._str_warnings()
        out += self._str_see_also(func_role)
        out += self._str_section('Notes')
        out += self._str_references()
        out += self._str_section('Examples')
        out = self._str_indent(out,indent)
        return '\n'.join(out)

class SphinxFunctionDoc(SphinxDocString, FunctionDoc):
    pass

class SphinxClassDoc(SphinxDocString, ClassDoc):
    pass

def get_doc_object(obj, what=None):
    if what is None:
        if inspect.isclass(obj):
            what = 'class'
        elif inspect.ismodule(obj):
            what = 'module'
        elif callable(obj):
            what = 'function'
        else:
            what = 'object'
    if what == 'class':
        return SphinxClassDoc(obj, '', func_doc=SphinxFunctionDoc)
    elif what in ('function', 'method'):
        return SphinxFunctionDoc(obj, '')
    else:
        return SphinxDocString(pydoc.getdoc(obj))



from .uriutil import check_entry
from .jsonutil import JsonTable

class Users(object):
    """ Database user management interface. It is used to retrieve information
        on users registered on a server.

        .. note::
        
            At the moment user creation and deletion is not supported through 
            the REST API but it will be at some point.

        Examples
        --------
            >>> interface.manage.users()
            ['list_of_users']
            >>> interface.manage.users.firstname('nosetests')
            'nose'

        See Also
        --------
        Project.users()
        Project.add_user()
        Project.remove_user()
    """

    def __init__(self, interface):
        """ 
            Parameters
            ----------
            interface: :class:`Interface`
                Main interface reference.
        """
        self._intf = interface

    def __call__(self):
        """ Returns the list of all registered users on the server.
        """
        self._intf._get_entry_point()

        return JsonTable(self._intf._get_json('%s/users' % self._intf._entry)
                         ).get('login', always_list=True)

    def firstname(self, login):
        """ Returns the firstname of the user.
        """
        self._intf._get_entry_point()

        return JsonTable(self._intf._get_json('%s/users' % self._intf._entry)
                         ).where(login=login)['firstname']

    def lastname(self, login):
        """ Returns the lastname of the user.
        """
        self._intf._get_entry_point()

        return JsonTable(self._intf._get_json('%s/users' % self._intf._entry)
                         ).where(login=login)['lastname']

    def id(self, login):
        """ Returns the id of the user.
        """
        self._intf._get_entry_point()

        return JsonTable(self._intf._get_json('%s/users' % self._intf._entry)
                         ).where(login=login)['xdat_user_id']

    def email(self, login):
        """ Returns the email of the user.
        """
        self._intf._get_entry_point()

        return JsonTable(self._intf._get_json('%s/users' % self._intf._entry)
                         ).where(login=login)['email']


    def resources(self):
        """ Returns the resources of the user.
        """
        self._intf._get_entry_point()

        print(self._intf._get_json(
            '%s/user/cache/resources' % self._intf._entry))



#!/usr/bin/env python

from distutils.core import setup
import sys,os

# For some commands, use setuptools
if len(set(['develop', 'sdist', 'release', 'bdist_egg', 'bdist_rpm', 'bdist',
            'bdist_dumb', 'bdist_wininst', 'install_egg_info', 'build_sphinx',
            'egg_info', 'easy_install', 'upload']).intersection(sys.argv)) > 0:
    from setupegg import extra_setuptools_args

# extra_setuptools_args is injected by the setupegg.py script, for
# running the setup with setuptools.
if not 'extra_setuptools_args' in globals():
    extra_setuptools_args = dict()

def get_version():
    basedir = os.path.dirname(__file__)
    with open(os.path.join(basedir, 'pyxnat/version.py')) as f:
        VERSION = None
        exec(f.read())
        return VERSION
    raise RuntimeError("No version found")

LONG_DESCRIPTION = """
PyXNAT
======

pyxnat provides an API to access data on XNAT (see http://xnat.org)
servers.

Visit http://packages.python.org/pyxnat for more information.
"""

setup(name='pyxnat',
      version=get_version(),
      summary='XNAT in Python',
      author='Yannick Schwartz',
      author_email='yannick.schwartz@cea.fr',
      url='http://packages.python.org/pyxnat/',
      packages=['pyxnat'],
      package_data={'pyxnat': ['externals/*.py',
                               'externals/httplib2/*.py',
                               'externals/simplejson/*.py',
                               'tests/*.py',
                               'tests/*.txt',
                               'tests/*.csv',
                               'core/*.py',
                               '*.py'],
                    },
      description="""Xnat in Python""",
      long_description=LONG_DESCRIPTION,
      license='BSD',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Education',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Scientific/Engineering',
          'Topic :: Utilities',
          'Topic :: Internet :: WWW/HTTP',
      ],

      platforms='any', requires=['httplib2'],
      **extra_setuptools_args)



import os
import socket
import platform
import tempfile
from uuid import uuid1

from .. import Interface

_modulepath = os.path.dirname(os.path.abspath(__file__))

central = Interface('https://central.xnat.org', 'nosetests', 'nosetests')

_id_set1 = {
    'sid':uuid1().hex,
    'eid':uuid1().hex,
    'aid':uuid1().hex,
    'cid':uuid1().hex,
    'rid':uuid1().hex,
    }

_id_set2 = {
    'sid':uuid1().hex,
    'eid':uuid1().hex,
    'aid':uuid1().hex,
    'cid':uuid1().hex,
    'rid':uuid1().hex,
    }

subj_1 = central.select.project('nosetests').subject(_id_set1['sid'])
expe_1 = subj_1.experiment(_id_set1['eid'])
asse_1 = expe_1.assessor(_id_set1['aid'])
scan_1 = expe_1.scan(_id_set1['cid'])
reco_1 = expe_1.reconstruction(_id_set1['rid'])

def test_subject_create():
    assert not subj_1.exists()
    subj_1.create()
    assert subj_1.exists()

def test_experiment_create():
    assert not expe_1.exists()
    expe_1.create()
    assert expe_1.exists()

def test_assessor_create():
    assert not asse_1.exists()
    asse_1.create()
    assert asse_1.exists()

def test_scan_create():
    assert not scan_1.exists()
    scan_1.create()
    assert scan_1.exists()

def test_reconstruction_create():
    assert not reco_1.exists()
    reco_1.create()
    assert reco_1.exists()

def test_provenance():
    reco_1.provenance.set({'program':'nosetests'})
    assert reco_1.provenance.get()[0]['program'] == 'nosetests'

def test_multi_create():
    asse_2 = central.select('/projects/nosetests/subjects/%(sid)s'
                            '/experiments/%(eid)s'
                            '/assessors/%(aid)s' % _id_set2
                            )

    expe_2 = central.select('/projects/nosetests/subjects/%(sid)s'
                            '/experiments/%(eid)s'%_id_set2
                            )

    assert not asse_2.exists()
    asse_2.create(experiments='xnat:petSessionData',                
                  assessors='xnat:petAssessorData')
    assert asse_2.exists()

    assert asse_2.datatype() == 'xnat:petAssessorData'
    assert expe_2.datatype() == 'xnat:petSessionData'

    scan_2 = central.select('/projects/nosetests/subjects/%(sid)s'
                            '/experiments/%(eid)s/scans/%(cid)s' % _id_set2
                            ).create()

    assert scan_2.datatype() == 'xnat:mrScanData'

#def test_share_subject():
#    assert not central.select('/projects/nosetests2'
#                              '/subjects/%(sid)s' % _id_set1
#                              ).exists()
#    subj_1.share('nosetests2')
#    assert central.select('/projects/nosetests2/subjects/%(sid)s'%_id_set1
#                              ).exists()
#    assert set(subj_1.shares().get()) == set(['nosetests', 'nosetests2'])

#def test_share_experiment():
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'
#                          '/experiments/%(eid)s'%_id_set1
#                          ).exists()
#    expe_1.share('nosetests2')
#    assert central.select('/projects/nosetests2/subjects/%(sid)s'
#                          '/experiments/%(eid)s'%_id_set1
#                          ).exists()
#    assert set(expe_1.shares().get()) == set(['nosetests', 'nosetests2'])

#def test_share_assessor():
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'
#                              '/experiments/%(eid)s/assessors/%(aid)s'%_id_set1
#                              ).exists()
#    asse_1.share('nosetests2')
#    assert central.select('/projects/nosetests2/subjects/%(sid)s'
#                          '/experiments/%(eid)s/assessors/%(aid)s'%_id_set1
#                          ).exists()
#    assert set(asse_1.shares().get()) == set(['nosetests', 'nosetests2'])

#def test_unshare_assessor():
#    asse_1.unshare('nosetests2')
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'
#                              '/experiments/%(eid)s/assessors/%(aid)s'%_id_set1
#                              ).exists()
#    assert asse_1.shares().get() == ['nosetests']

#def test_unshare_experiment():
#    expe_1.unshare('nosetests2')
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'
#                          '/experiments/%(eid)s'%_id_set1
#                          ).exists()
#    assert expe_1.shares().get() == ['nosetests']

#def test_unshare_subject():
#    subj_1.unshare('nosetests2')
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'%_id_set1
#                              ).exists()
#    assert subj_1.shares().get() == ['nosetests']

def test_put_file():
    local_path = os.path.join(_modulepath, 'hello_xnat.txt')
    subj_1.resource('test').file('hello.txt').put(local_path)
    assert subj_1.resource('test').file('hello.txt').exists()
    assert long(subj_1.resource('test').file('hello.txt').size()) == \
                                                os.stat(local_path).st_size

def test_get_file():
    fh = subj_1.resource('test').file('hello.txt')

    central.cache.set_usage(expiration=0)

    fpath = fh.get()
    assert os.path.exists(fpath)
    assert open(fpath, 'rb').read() == 'Hello XNAT!\n'

    custom = os.path.join(tempfile.gettempdir(), uuid1().hex)
    
    fh.get(custom)
    assert os.path.exists(custom)
    assert not os.path.exists(fpath)
    fh.get()
    assert os.path.exists(custom)
    os.remove(custom)

def test_put_dir_file():
    local_path = os.path.join(_modulepath, 'hello_again.txt')
    subj_1.resource('test').file('dir/hello.txt').put(local_path)
    assert subj_1.resource('test').file('dir/hello.txt').exists()
    assert long(subj_1.resource('test').file('dir/hello.txt').size()) == \
                                                os.stat(local_path).st_size

def test_get_dir_file():
    fh = subj_1.resource('test').file('dir/hello.txt')

    central.cache.set_usage(expiration=0)

    fpath = fh.get()
    assert os.path.exists(fpath)
    assert open(fpath, 'rb').read() == 'Hello again!\n'

    custom = os.path.join(tempfile.gettempdir(), uuid1().hex)
    
    fh.get(custom)
    assert os.path.exists(custom)
    assert not os.path.exists(fpath)
    fh.get()
    assert os.path.exists(custom)
    os.remove(custom)

def test_get_copy_file():
    fpath = os.path.join(tempfile.gettempdir(), uuid1().hex)
    fpath = subj_1.resource('test').file('hello.txt').get_copy(fpath)
    assert os.path.exists(fpath)
    fd = open(fpath, 'rb')
    assert fd.read() == 'Hello XNAT!\n'
    fd.close()
    os.remove(fpath)

    central.cache.set_usage(expiration=1)

def test_file_last_modified():
    f = subj_1.resource('test').file('hello.txt')
    assert isinstance(f.last_modified(), basestring)
    assert len(f.last_modified()) > 0

def test_last_modified():
    sid = subj_1.id()

    t1 = central.select('/project/nosetests').last_modified()[sid]
    subj_1.attrs.set('age', '26')
    assert subj_1.attrs.get('age') == '26'
    t2 = central.select('/project/nosetests').last_modified()[sid]

    assert t1 != t2

def test_getitem_key():
    projects = central.select.projects()
    assert projects.first().id() == projects[0].id()
    piter = projects.__iter__()
    next(piter)
    assert next(piter).id() == projects[1].id()

def test_getitem_slice():
    projects = central.select.projects()
    assert projects.first().id() == next(projects[:1]).id()
    piter = projects.__iter__()
    next(piter)
    next(piter)
    next(piter)
    for pobj in projects[3:6]:
        assert next(piter).id() == pobj.id()

def test_subject1_parent():
    project = central.select.project('nosetests')
    assert subj_1.parent()._uri == project._uri

def test_project_parent():
    project = central.select.project('nosetests')
    assert not project.parent()

def test_subject1_delete():
    assert subj_1.exists()
    subj_1.delete()
    assert not subj_1.exists()

def test_subject2_delete():
    subj_2 = central.select('/projects/nosetests/subjects/%(sid)s'%_id_set2)
    assert subj_2.exists()
    subj_2.delete()
    assert not subj_2.exists()

def test_project_configuration():
    project = central.select('/project/nosetests')
    assert project.quarantine_code() == 0
    assert project.prearchive_code() == 0
    assert project.current_arc() == 'arc001'
    assert 'nosetests' in project.users()
    assert 'nosetests' in project.owners()
    assert project.user_role('nosetests') == 'owner'                              
                                              
def test_put_zip():
    local_path = os.path.join(_modulepath, 'hello_dir.zip')
    assert os.path.exists(local_path)

    # Upload and confirm proper extraction
    r1 = subj_1.resource('test_zip_extract')
    r1.put_zip(local_path, extract=True)
    assert r1.exists()
    assert r1.file('hello_dir/hello_xnat_dir.txt').exists()
    assert r1.file('hello_dir/hello_dir2/hello_xnat_dir2.txt').exists()
   
    # Upload and confirm not extracted 
    r2 = subj_1.resource('test_zip_no_extract')
    r2.put_zip(local_path, extract=False)
    assert r2.exists()
    assert r2.file('hello_dir.zip').exists()
                                                                                                
def test_get_zip():
    r = subj_1.resource('test_zip_extract')
    local_dir = os.path.join(_modulepath, 'test_zip_download'+r.id())
    file_list = [os.path.join(local_dir,'test_zip_extract/hello_dir'), 
                 os.path.join(local_dir,'test_zip_extract/hello_dir/hello_xnat_dir.txt'),
                 os.path.join(local_dir,'test_zip_extract/hello_dir/hello_dir2'),
                 os.path.join(local_dir,'test_zip_extract/hello_dir/hello_dir2/hello_xnat_dir2.txt')]

    if not os.path.exists(local_dir):
        os.mkdir(local_dir)
        
    r.get(local_dir, extract=True)
    for f in file_list:
        assert os.path.exists(f)
